package main

import (
	"context"
	"fmt"
	"net"
	"net/http"
	"os"

	"github.com/jmoiron/sqlx"
	_ "github.com/lib/pq" // Postgres driver initialization
	account "gitlab.mashroom.com/mashroom/backend/account/api"
	auth "gitlab.mashroom.com/mashroom/backend/auth/api"
	"gitlab.mashroom.com/mashroom/backend/auth/internal/model/postgres"
	"gitlab.mashroom.com/mashroom/backend/auth/internal/service"
	"gitlab.mashroom.com/mashroom/backend/auth/internal/storage/keycloak"
	"gitlab.mashroom.com/mashroom/backend/auth/version"
	"gitlab.mashroom.com/mashroom/backend/common/api/health"
	"gitlab.mashroom.com/mashroom/backend/common/generic/addr"
	"gitlab.mashroom.com/mashroom/backend/common/generic/env"
	"gitlab.mashroom.com/mashroom/backend/common/generic/error/sqlerr"
	"gitlab.mashroom.com/mashroom/backend/common/generic/grpc"
	logger "gitlab.mashroom.com/mashroom/backend/common/generic/log"
	"gitlab.mashroom.com/mashroom/backend/common/generic/metrics"
	"gitlab.mashroom.com/mashroom/backend/common/generic/tracer"
	invitation "gitlab.mashroom.com/mashroom/backend/invitation/api"
	notification "gitlab.mashroom.com/mashroom/backend/notification/api"
)

var (
	envKeycloakURL          = env.ResolveString("KEYCLOAK_URL", "http://keycloak:8080/auth")
	envKeycloakClient       = env.ResolveString("KEYCLOAK_CLIENT", "mashroom")
	envKeycloakClientSecret = env.ResolveString("KEYCLOAK_SECRET", "")
	envKeycloakRealm        = env.ResolveString("KEYCLOAK_REALM", "master")
)

var (
	envAuthThreshold = env.ResolveInt("AUTH_THRESHOLD", 10)
	envRecaptchaKey  = env.ResolveString("RECAPTCHA_KEY", "")
	envOAuthEnabled  = env.ResolveBool("OAUTH_ENABLED", false)
)

var accountService = account.NewAccountServiceClient(grpc.Dial(
	env.ResolveString("ACCOUNT_SERVICE_URL", "mashroom-account:10000"),
	"Account",
))

var notificationService = notification.NewNotificationServiceClient(grpc.Dial(
	env.ResolveString("NOTIFICATION_SERVICE_URL", "mashroom-notification:10000"),
	"Notification",
))

var invitationService = invitation.NewInvitationServiceClient(grpc.Dial(
	env.ResolveString("INVITATION_SERVICE_URL", "mashroom-invitation:10000"),
	"Invitation",
))

func main() {
	ctx := context.Background()

	defer tracer.Close()

	conn := fmt.Sprintf(
		"postgres://%s:%s@%s:%s/%s?sslmode=%s",
		os.Getenv("POSTGRES_USERNAME"),
		os.Getenv("POSTGRES_PASSWORD"),
		os.Getenv("POSTGRES_HOST"),
		os.Getenv("POSTGRES_PORT"),
		os.Getenv("KEYCLOAK_DB"),
		os.Getenv("POSTGRES_SSL_MODE"),
	)

	db := sqlx.MustOpen("postgres", conn)

	repo, err := postgres.NewRepository(db)
	if err != nil {
		panic(err)
	}

	storage, err := keycloak.NewRepository(ctx, &keycloak.Config{
		URL:        envKeycloakURL,
		Client:     envKeycloakClient,
		Secret:     envKeycloakClientSecret,
		Realm:      envKeycloakRealm,
		HTTPClient: &http.Client{},
		Debug:      os.Getenv(logger.LoglevelEnv) == logger.LoglevelDebug,
	})
	if err != nil {
		panic(err)
	}

	connections := service.Connections{
		Account:      accountService,
		Notification: notificationService,
		Invitation:   invitationService,
	}

	srv, err := service.NewService(&service.Config{
		Repo:          repo,
		Storage:       storage,
		Connections:   connections,
		AuthThreshold: envAuthThreshold,
		ReCaptchaKey:  envRecaptchaKey,
		OAuthEnabled:  envOAuthEnabled,
	})
	if err != nil {
		panic(err)
	}

	listener, err := net.Listen("tcp", addr.Grpc)
	if err != nil {
		panic(err)
	}

	grpcServer, err := grpc.NewServer("auth", grpc.ServerOptionErrorTranslator(
		sqlerr.Translate,
	))
	if err != nil {
		panic(err)
	}

	auth.RegisterAuthServiceServer(grpcServer, srv)

	logger.Infoln("auth started")

	healthServer := version.NewHealthServer()
	health.RegisterHealthServiceServer(grpcServer, healthServer)

	go func() {
		_ = metrics.ConfigureListenAndServe(grpcServer)
	}()

	err = grpcServer.Serve(listener)
	if err != nil {
		logger.Errorln("gRPC error", err)
	}
}
