package test

import (
	"github.com/golang/mock/gomock"
	account "gitlab.mashroom.com/mashroom/backend/account/api"
	auth "gitlab.mashroom.com/mashroom/backend/auth/api"
	"gitlab.mashroom.com/mashroom/backend/auth/internal/model"
	"gitlab.mashroom.com/mashroom/backend/auth/internal/storage"
	"gitlab.mashroom.com/mashroom/backend/common/api/grpcerr"
)

func (s *serviceSuite) TestAuthLogin_ok() {
	s.account.EXPECT().GetByEmail(s.testCtx, &account.GetByEmailRequest{
		Email: s.expectUserEmail(),
		Admin: false,
	}).Return(&account.GetResponse{
		Account: &account.Account{
			Id: &account.Account_IdValue{IdValue: s.expectUserID()},
			Contact: &account.Account_ContactValue{
				ContactValue: &account.AccountContact{
					EmailPrimary:    &account.AccountContact_EmailPrimaryValue{EmailPrimaryValue: s.expectUserEmail()},
					PhonePrimary:    nil,
					PhonesSecondary: nil,
					ContactStatus:   nil,
				},
			},
		},
	}, nil)

	s.repo.EXPECT().AuthAttemptInsert(s.testCtx, &model.AuthAttempt{
		Remote: s.expectUserRemote(),
	}).Return(nil)

	s.repo.EXPECT().AuthAttemptCountByRemote(s.testCtx, s.expectUserRemote()).Return(1, nil)

	s.storage.EXPECT().SessionLogin(s.testCtx, s.expectUserEmail(), "efg").Return(s.expectSessionAuth(0), nil)

	s.expectSessionIntrospect(s.expectAccessToken(0), true)

	s.repo.EXPECT().AuthAttemptRemoveByRemote(s.testCtx, s.expectUserRemote()).Return(nil)

	s.account.EXPECT().UpdateLastLogin(s.testCtx, &account.UpdateLastLoginRequest{
		Id: s.expectUserID(),
	}).Return(&account.UpdateLastLoginResponse{}, nil)

	s.repo.EXPECT().RecordUserLogin(s.testCtx, s.expectUserID(), s.expectUserRemote()).Return(false, true, nil)

	resp, err := s.srv.Login(s.testCtx, &auth.LoginRequest{
		Email:       &auth.LoginRequest_EmailValue{EmailValue: s.expectUserEmail()},
		Password:    &auth.LoginRequest_PasswordValue{PasswordValue: "efg"},
		TosAccepted: nil,
		PhoneOtp:    nil,
		EmailOtp:    nil,
		Remote:      &auth.LoginRequest_RemoteValue{RemoteValue: s.expectUserRemote()},
		Captcha:     nil,
		AdminSite:   nil,
	})

	s.Require().NoError(err)

	s.Require().NotNil(resp)

	s.Require().Empty(resp.GetErrors())
}

func (s *serviceSuite) TestAuthLogin_ok_unknown() {
	s.account.EXPECT().GetByEmail(s.testCtx, &account.GetByEmailRequest{
		Email: s.expectUserEmail(),
		Admin: false,
	}).Return(&account.GetResponse{
		Account: &account.Account{
			Id: &account.Account_IdValue{IdValue: s.expectUserID()},
			Contact: &account.Account_ContactValue{
				ContactValue: &account.AccountContact{
					EmailPrimary:    &account.AccountContact_EmailPrimaryValue{EmailPrimaryValue: s.expectUserEmail()},
					PhonePrimary:    nil,
					PhonesSecondary: nil,
					ContactStatus:   nil,
				},
			},
		},
	}, nil)

	s.repo.EXPECT().AuthAttemptInsert(s.testCtx, &model.AuthAttempt{
		Remote: s.expectUserRemote(),
	}).Return(nil)

	s.repo.EXPECT().AuthAttemptCountByRemote(s.testCtx, s.expectUserRemote()).Return(1, nil)

	s.storage.EXPECT().SessionLogin(s.testCtx, s.expectUserEmail(), "efg").Return(s.expectSessionAuth(0), nil)

	s.expectSessionIntrospect(s.expectAccessToken(0), true)

	s.repo.EXPECT().AuthAttemptRemoveByRemote(s.testCtx, s.expectUserRemote()).Return(nil)

	s.account.EXPECT().UpdateLastLogin(s.testCtx, &account.UpdateLastLoginRequest{
		Id: s.expectUserID(),
	}).Return(&account.UpdateLastLoginResponse{}, nil)

	s.repo.EXPECT().RecordUserLogin(s.testCtx, s.expectUserID(), s.expectUserRemote()).Return(false, false, nil)

	s.notification.EXPECT().SendNotification(s.testCtx, gomock.Any()).Return(nil, nil)

	resp, err := s.srv.Login(s.testCtx, &auth.LoginRequest{
		Email:       &auth.LoginRequest_EmailValue{EmailValue: s.expectUserEmail()},
		Password:    &auth.LoginRequest_PasswordValue{PasswordValue: "efg"},
		TosAccepted: nil,
		PhoneOtp:    nil,
		EmailOtp:    nil,
		Remote:      &auth.LoginRequest_RemoteValue{RemoteValue: s.expectUserRemote()},
		Captcha:     nil,
		AdminSite:   nil,
	})

	s.Require().NoError(err)

	s.Require().NotNil(resp)

	s.Require().Empty(resp.GetErrors())
}

func (s *serviceSuite) TestAuthLogin_error_tos() {
	s.account.EXPECT().GetByEmail(s.testCtx, &account.GetByEmailRequest{
		Email: s.expectUserEmail(),
		Admin: false,
	}).Return(&account.GetResponse{
		Account: &account.Account{
			Id: &account.Account_IdValue{IdValue: s.expectUserID()},
			Contact: &account.Account_ContactValue{
				ContactValue: &account.AccountContact{
					EmailPrimary:    &account.AccountContact_EmailPrimaryValue{EmailPrimaryValue: s.expectUserEmail()},
					PhonePrimary:    nil,
					PhonesSecondary: nil,
					ContactStatus:   nil,
				},
			},
		},
	}, nil)

	s.repo.EXPECT().AuthAttemptInsert(s.testCtx, &model.AuthAttempt{
		Remote: s.expectUserRemote(),
	}).Return(nil)

	s.repo.EXPECT().AuthAttemptCountByRemote(s.testCtx, s.expectUserRemote()).Return(1, nil)

	s.storage.EXPECT().SessionLogin(s.testCtx, s.expectUserEmail(), "efg").Return(s.expectSessionAuth(0), nil)

	s.expectSessionIntrospect(s.expectAccessToken(0), true, "tosupdate")

	resp, err := s.srv.Login(s.testCtx, &auth.LoginRequest{
		Email:       &auth.LoginRequest_EmailValue{EmailValue: s.expectUserEmail()},
		Password:    &auth.LoginRequest_PasswordValue{PasswordValue: "efg"},
		TosAccepted: nil,
		PhoneOtp:    nil,
		EmailOtp:    nil,
		Remote:      &auth.LoginRequest_RemoteValue{RemoteValue: s.expectUserRemote()},
		Captcha:     nil,
		AdminSite:   nil,
	})

	s.Require().NoError(err)

	s.Require().Contains(resp.Errors, auth.AuthErrors_AUTH_ERROR_TOS_REVIEW_REQUIRED)
}

func (s *serviceSuite) TestAuthLogin_error_admin() {
	s.account.EXPECT().GetByEmail(s.testCtx, &account.GetByEmailRequest{
		Email: s.expectUserEmail(),
		Admin: false,
	}).Return(&account.GetResponse{
		Account: &account.Account{
			Id: &account.Account_IdValue{IdValue: s.expectUserID()},
			Contact: &account.Account_ContactValue{
				ContactValue: &account.AccountContact{
					EmailPrimary:    &account.AccountContact_EmailPrimaryValue{EmailPrimaryValue: s.expectUserEmail()},
					PhonePrimary:    nil,
					PhonesSecondary: nil,
					ContactStatus:   nil,
				},
			},
		},
	}, nil)

	s.repo.EXPECT().AuthAttemptInsert(s.testCtx, &model.AuthAttempt{
		Remote: s.expectUserRemote(),
	}).Return(nil)

	s.repo.EXPECT().AuthAttemptCountByRemote(s.testCtx, s.expectUserRemote()).Return(1, nil)

	s.storage.EXPECT().SessionLogin(s.testCtx, s.expectUserEmail(), "efg").Return(s.expectSessionAuth(0), nil)

	s.expectSessionIntrospect(s.expectAccessToken(0), true, "admin")

	_, err := s.srv.Login(s.testCtx, &auth.LoginRequest{
		Email:       &auth.LoginRequest_EmailValue{EmailValue: s.expectUserEmail()},
		Password:    &auth.LoginRequest_PasswordValue{PasswordValue: "efg"},
		TosAccepted: nil,
		PhoneOtp:    nil,
		EmailOtp:    nil,
		Remote:      &auth.LoginRequest_RemoteValue{RemoteValue: s.expectUserRemote()},
		Captcha:     nil,
		AdminSite:   nil,
	})

	s.Require().ErrorIs(err, grpcerr.NoAuth)
}

func (s *serviceSuite) TestAuthLogin_error_introspect() {
	s.account.EXPECT().GetByEmail(s.testCtx, &account.GetByEmailRequest{
		Email: s.expectUserEmail(),
		Admin: false,
	}).Return(&account.GetResponse{
		Account: &account.Account{
			Id: &account.Account_IdValue{IdValue: s.expectUserID()},
			Contact: &account.Account_ContactValue{
				ContactValue: &account.AccountContact{
					EmailPrimary:    &account.AccountContact_EmailPrimaryValue{EmailPrimaryValue: s.expectUserEmail()},
					PhonePrimary:    nil,
					PhonesSecondary: nil,
					ContactStatus:   nil,
				},
			},
		},
	}, nil)

	s.repo.EXPECT().AuthAttemptInsert(s.testCtx, &model.AuthAttempt{
		Remote: s.expectUserRemote(),
	}).Return(nil)

	s.repo.EXPECT().AuthAttemptCountByRemote(s.testCtx, s.expectUserRemote()).Return(1, nil)

	s.storage.EXPECT().SessionLogin(s.testCtx, s.expectUserEmail(), "efg").Return(s.expectSessionAuth(0), nil)

	s.storage.EXPECT().SessionIntrospect(s.testCtx, s.expectAccessToken(0)).
		Return(nil, s.expectError())

	_, err := s.srv.Login(s.testCtx, &auth.LoginRequest{
		Email:       &auth.LoginRequest_EmailValue{EmailValue: s.expectUserEmail()},
		Password:    &auth.LoginRequest_PasswordValue{PasswordValue: "efg"},
		TosAccepted: nil,
		PhoneOtp:    nil,
		EmailOtp:    nil,
		Remote:      &auth.LoginRequest_RemoteValue{RemoteValue: s.expectUserRemote()},
		Captcha:     nil,
		AdminSite:   nil,
	})

	s.Require().Error(err)
}

func (s *serviceSuite) TestAuthLogin_error_login() {
	s.account.EXPECT().GetByEmail(s.testCtx, &account.GetByEmailRequest{
		Email: s.expectUserEmail(),
		Admin: false,
	}).Return(&account.GetResponse{
		Account: &account.Account{
			Id: &account.Account_IdValue{IdValue: s.expectUserID()},
			Contact: &account.Account_ContactValue{
				ContactValue: &account.AccountContact{
					EmailPrimary:    &account.AccountContact_EmailPrimaryValue{EmailPrimaryValue: s.expectUserEmail()},
					PhonePrimary:    nil,
					PhonesSecondary: nil,
					ContactStatus:   nil,
				},
			},
		},
	}, nil)

	s.repo.EXPECT().AuthAttemptInsert(s.testCtx, &model.AuthAttempt{
		Remote: s.expectUserRemote(),
	}).Return(nil)

	s.repo.EXPECT().AuthAttemptCountByRemote(s.testCtx, s.expectUserRemote()).Return(1, nil)

	s.storage.EXPECT().SessionLogin(s.testCtx, s.expectUserEmail(), "efg").Return(nil, s.expectError())

	resp, err := s.srv.Login(s.testCtx, &auth.LoginRequest{
		Email:       &auth.LoginRequest_EmailValue{EmailValue: s.expectUserEmail()},
		Password:    &auth.LoginRequest_PasswordValue{PasswordValue: "efg"},
		TosAccepted: nil,
		PhoneOtp:    nil,
		EmailOtp:    nil,
		Remote:      &auth.LoginRequest_RemoteValue{RemoteValue: s.expectUserRemote()},
		Captcha:     nil,
		AdminSite:   nil,
	})

	s.Require().NoError(err)

	s.Require().Contains(resp.Errors, auth.AuthErrors_AUTH_ERROR_INVALID_PASSWORD)
}

func (s *serviceSuite) TestAuthLogin_error_attempts() {
	s.account.EXPECT().GetByEmail(s.testCtx, &account.GetByEmailRequest{
		Email: s.expectUserEmail(),
		Admin: false,
	}).Return(&account.GetResponse{
		Account: &account.Account{
			Id: &account.Account_IdValue{IdValue: s.expectUserID()},
			Contact: &account.Account_ContactValue{
				ContactValue: &account.AccountContact{
					EmailPrimary:    &account.AccountContact_EmailPrimaryValue{EmailPrimaryValue: s.expectUserEmail()},
					PhonePrimary:    nil,
					PhonesSecondary: nil,
					ContactStatus:   nil,
				},
			},
		},
	}, nil)

	s.repo.EXPECT().AuthAttemptInsert(s.testCtx, &model.AuthAttempt{
		Remote: s.expectUserRemote(),
	}).Return(nil)

	s.repo.EXPECT().AuthAttemptCountByRemote(s.testCtx, s.expectUserRemote()).Return(10, nil)

	s.storage.EXPECT().UserGetByID(s.testCtx, s.expectUserID()).Return(&storage.User{
		Enabled: true,
		Id:      s.expectUserID(),
	}, nil)

	s.storage.EXPECT().UserUpdate(s.testCtx, &storage.User{
		Enabled: false,
		Id:      s.expectUserID(),
	}).Return(nil)

	s.account.EXPECT().Block(s.testCtx, &account.BlockRequest{
		Id:    s.expectUserID(),
		Block: true,
	}).Return(&account.BlockResponse{}, nil)

	s.repo.EXPECT().AuthAttemptRemoveByRemote(s.testCtx, s.expectUserRemote()).Return(nil)

	resp, err := s.srv.Login(s.testCtx, &auth.LoginRequest{
		Email:       &auth.LoginRequest_EmailValue{EmailValue: s.expectUserEmail()},
		Password:    &auth.LoginRequest_PasswordValue{PasswordValue: "efg"},
		TosAccepted: nil,
		PhoneOtp:    nil,
		EmailOtp:    nil,
		Remote:      &auth.LoginRequest_RemoteValue{RemoteValue: s.expectUserRemote()},
		Captcha:     nil,
		AdminSite:   nil,
	})

	s.Require().NoError(err)

	s.Require().Contains(resp.Errors, auth.AuthErrors_AUTH_ERROR_ACCOUNT_IS_BLOCKED)
}

func (s *serviceSuite) TestAuthLogin_error_get() {
	s.account.EXPECT().GetByEmail(s.testCtx, &account.GetByEmailRequest{
		Email: s.expectUserEmail(),
		Admin: false,
	}).Return(nil, grpcerr.NotFound)

	resp, err := s.srv.Login(s.testCtx, &auth.LoginRequest{
		Email:       &auth.LoginRequest_EmailValue{EmailValue: s.expectUserEmail()},
		Password:    &auth.LoginRequest_PasswordValue{PasswordValue: "efg"},
		TosAccepted: nil,
		PhoneOtp:    nil,
		EmailOtp:    nil,
		Remote:      &auth.LoginRequest_RemoteValue{RemoteValue: s.expectUserRemote()},
		Captcha:     nil,
		AdminSite:   nil,
	})

	s.Require().NoError(err)

	s.Require().Contains(resp.Errors, auth.AuthErrors_AUTH_ERROR_ACCOUNT_DOES_NOT_EXIST)
}

func (s *serviceSuite) TestAuthLogin_error_blocked() {
	s.account.EXPECT().GetByEmail(s.testCtx, &account.GetByEmailRequest{
		Email: s.expectUserEmail(),
		Admin: false,
	}).Return(&account.GetResponse{
		Account: &account.Account{
			Id: &account.Account_IdValue{IdValue: s.expectUserID()},
			Contact: &account.Account_ContactValue{
				ContactValue: &account.AccountContact{
					EmailPrimary:    &account.AccountContact_EmailPrimaryValue{EmailPrimaryValue: s.expectUserEmail()},
					PhonePrimary:    nil,
					PhonesSecondary: nil,
					ContactStatus:   nil,
				},
			},
			Block: &account.Account_BlockValue{BlockValue: true},
		},
	}, nil)

	resp, err := s.srv.Login(s.testCtx, &auth.LoginRequest{
		Email:       &auth.LoginRequest_EmailValue{EmailValue: s.expectUserEmail()},
		Password:    &auth.LoginRequest_PasswordValue{PasswordValue: "efg"},
		TosAccepted: nil,
		PhoneOtp:    nil,
		EmailOtp:    nil,
		Remote:      &auth.LoginRequest_RemoteValue{RemoteValue: s.expectUserRemote()},
		Captcha:     nil,
		AdminSite:   nil,
	})

	s.Require().NoError(err)

	s.Require().NotNil(resp)

	s.Require().Contains(resp.Errors, auth.AuthErrors_AUTH_ERROR_ACCOUNT_IS_BLOCKED)
}
