// Package keycloak provides keyclaok implementation for user persistent repository
package keycloak

import (
	"context"
	"encoding/base64"
	"errors"
	"net/http"
	"net/url"
	"strings"
	"sync"

	"gitlab.mashroom.com/mashroom/backend/auth/internal/storage"
	"gitlab.mashroom.com/mashroom/backend/auth/keycloak"
	"gitlab.mashroom.com/mashroom/backend/common/api/grpcerr"
)

var errUserLogin = grpcerr.New(grpcerr.NoAuth, "user login error")

// Config provides keycloak connection configuration
type Config struct {
	HTTPClient *http.Client
	URL        string
	Client     string
	Secret     string
	Realm      string
	Debug      bool
}

type repositoryImpl struct {
	adminMutex       *sync.RWMutex
	basicAuth        string
	adminAccessToken string
	openIDConfig     keycloak.OpenIDConfigurationRepresentation
	config           Config
}

// NewRepository returns new keycloak user repository
func NewRepository(ctx context.Context, config *Config) (repos storage.Repository, err error) {
	repo := &repositoryImpl{
		config:     *config,
		basicAuth:  "basic " + base64.StdEncoding.EncodeToString([]byte(config.Client+":"+config.Secret)),
		adminMutex: &sync.RWMutex{},
	}

	err = repo.clientGet(ctx, "/realms/"+config.Realm+"/.well-known/openid-configuration", &repo.openIDConfig)
	if err != nil {
		return nil, err
	}

	err = repo.adminTokenMaintain(ctx)
	if err != nil {
		return nil, err
	}

	err = repo.Migrate(ctx)
	if err != nil {
		return nil, err
	}

	return repo, nil
}

func (repo *repositoryImpl) Migrate(ctx context.Context) error {
	var scopes []keycloak.ClientScopeRepresentation

	err := repo.adminGet(ctx, "/admin/realms/"+repo.config.Realm+"/client-scopes", &scopes)
	if err != nil {
		return err
	}

	for _, s := range scopes {
		if s.Name != "profile" {
			continue
		}

		found := false

		for _, m := range s.ProtocolMappers {
			name, ok := m["name"].(string)
			if ok && name == "transient_session" {
				found = true
				break
			}
		}

		if found {
			continue
		}

		err = repo.adminPost(
			ctx,
			"/admin/realms/"+repo.config.Realm+"/client-scopes/"+s.Id+"/protocol-mappers/models",
			&keycloak.ProtocolMapperRepresentation{
				Config: map[string]interface{}{
					"usermodel.realmRoleMapping.rolePrefix": "",
					"multivalued":                           "",
					"claim.name":                            "transient_session",
					"jsonType.label":                        "String",
					"id.token.claim":                        "true",
					"access.token.claim":                    "true",
					"userinfo.token.claim":                  "true",
					"user.attribute":                        "transient_session",
				},
				Name:           "transient_session",
				Protocol:       "openid-connect",
				ProtocolMapper: "oidc-usermodel-attribute-mapper",
			},
		)
		if err != nil {
			return err
		}

		break
	}

	return nil
}

func (repo *repositoryImpl) SessionRefresh(
	ctx context.Context,
	refreshToken string,
) (*storage.SessionAuth, error) {
	auth := &storage.SessionAuth{}

	err := repo.clientPostForm(ctx, repo.openIDConfig.TokenEndpoint, url.Values{
		"client_id":     {repo.config.Client},
		"grant_type":    {"refresh_token"},
		"refresh_token": {refreshToken},
	}, auth)
	if err != nil {
		return nil, err
	}

	if auth.Error != "" {
		return nil, errors.New(auth.Error)
	}

	return auth, nil
}

func (repo *repositoryImpl) SessionLogin(
	ctx context.Context,
	user, password string,
) (*storage.SessionAuth, error) {
	var users []*keycloak.UserRepresentation

	err := repo.adminGet(ctx, "/admin/realms/"+repo.config.Realm+"/users?email="+url.QueryEscape(user), &users)

	if err != nil || len(users) == 0 {
		return nil, errUserLogin
	}

	auth := &storage.SessionAuth{}

	err = repo.clientPostForm(ctx, repo.openIDConfig.TokenEndpoint, url.Values{
		"client_id":  {repo.config.Client},
		"grant_type": {"password"},
		"username":   {users[0].Username},
		"password":   {password},
		"scope":      {"email phone profile"},
	}, auth)
	if err != nil {
		return nil, err
	}

	if auth.Error != "" {
		return nil, errors.New(auth.Error)
	}

	return auth, nil
}

func (repo *repositoryImpl) SessionIntrospect(
	ctx context.Context,
	accessToken string,
) (*storage.SessionIntrospection, error) {
	introspect := &storage.SessionIntrospection{}

	err := repo.clientPostForm(ctx, repo.openIDConfig.TokenIntrospectionEndpoint, url.Values{
		"token": {accessToken},
	}, introspect)
	if err != nil {
		return nil, err
	}

	if introspect.Error != "" {
		return nil, errors.New(introspect.Error)
	}

	return introspect, nil
}

func (repo *repositoryImpl) SessionImpersonate(
	ctx context.Context,
	userID string,
) (*storage.SessionAuth, error) {
	auth := &storage.SessionAuth{}

	err := repo.httpRetry(func() error {
		return repo.clientPostForm(ctx, repo.openIDConfig.TokenEndpoint, url.Values{
			"client_id":         {repo.config.Client},
			"grant_type":        {"urn:ietf:params:oauth:grant-type:token-exchange"},
			"subject_token":     {repo.adminToken()},
			"requested_subject": {userID},
		}, auth)
	})
	if err != nil {
		return nil, err
	}

	if auth.Error != "" {
		return nil, errors.New(auth.Error)
	}

	return auth, nil
}

func (repo *repositoryImpl) SessionRemove(ctx context.Context, sessionID string) error {
	return repo.adminDelete(ctx, "/admin/realms/"+repo.config.Realm+"/sessions/"+sessionID, nil)
}

func (repo *repositoryImpl) SessionRemoveAll(ctx context.Context, userID string) error {
	return repo.adminPost(
		ctx,
		"/admin/realms/"+repo.config.Realm+"/users/"+userID+"/logout",
		nil,
	)
}

func (repo *repositoryImpl) UserCreate(ctx context.Context, user *storage.User) error {
	return repo.adminPost(ctx, "/admin/realms/"+repo.config.Realm+"/users", user)
}

func (repo *repositoryImpl) UserUpdate(ctx context.Context, user *storage.User) error {
	return repo.adminPut(ctx, "/admin/realms/"+repo.config.Realm+"/users/"+user.Id, user, nil)
}

func (repo *repositoryImpl) UserUpdateRoles(ctx context.Context, userID string, roles []*storage.Role) error {
	return repo.adminPost(
		ctx,
		"/admin/realms/"+repo.config.Realm+"/users/"+userID+"/role-mappings/realm",
		roles,
	)
}

func (repo *repositoryImpl) UserPasswordReset(ctx context.Context, userID, password string) error {
	cred := map[string]interface{}{
		"type":      "password",
		"value":     password,
		"temporary": false,
	}

	return repo.adminPut(
		ctx,
		"/admin/realms/"+repo.config.Realm+"/users/"+userID+"/reset-password",
		cred,
		nil,
	)
}

func (repo *repositoryImpl) UserGetByID(ctx context.Context, userID string) (user *storage.User, err error) {
	user = &storage.User{}
	err = repo.adminGet(ctx, "/admin/realms/"+repo.config.Realm+"/users/"+userID, user)

	return
}

func (repo *repositoryImpl) UserGetByEmail(ctx context.Context, email string) (ms []*storage.User, err error) {
	var users []*storage.User

	email = strings.ToLower(email)

	err = repo.adminGet(ctx, "/admin/realms/"+repo.config.Realm+"/users?email="+url.QueryEscape(email), &users)
	if err != nil {
		return nil, err
	}

	for _, u := range users {
		if u.Email == email {
			ms = append(ms, u)
		}
	}

	return
}

func (repo *repositoryImpl) RoleGetByCode(ctx context.Context, code string) (role *storage.Role, err error) {
	role = &storage.Role{}
	err = repo.adminGet(ctx, "/admin/realms/"+repo.config.Realm+"/roles/"+url.PathEscape(code), role)

	return
}
