package keycloak

import (
	"bytes"
	"context"
	"encoding/json"
	"errors"
	"fmt"
	"io"
	"io/ioutil"
	"net/http"
	"net/http/httputil"
	"net/url"
	"os"
	"strings"
	"time"

	"gitlab.mashroom.com/mashroom/backend/auth/keycloak"
	"gitlab.mashroom.com/mashroom/backend/common/generic/log"
)

const err401Unauthorized = "HTTP 401 Unauthorized"

const httpRetries = 1

type adminAuthResponse struct {
	keycloak.OpenIDAuthRepresentation
	keycloak.OpenIDErrorRepresentation
}

func (repo *repositoryImpl) adminTokenFetch(ctx context.Context) (expire int, access string, err error) {
	var resp adminAuthResponse

	err = repo.clientPostForm(ctx, repo.openIDConfig.TokenEndpoint, url.Values{
		"client_id":  {repo.config.Client},
		"grant_type": {"client_credentials"},
	}, &resp)
	if err != nil {
		return
	}

	if resp.Error != "" {
		err = errors.New(resp.Error)
		return
	}

	return resp.ExpiresIn, resp.AccessToken, nil
}

func (repo *repositoryImpl) adminToken() string {
	repo.adminMutex.RLock()
	defer repo.adminMutex.RUnlock()

	return repo.adminAccessToken
}

func (repo *repositoryImpl) adminTokenUpdate(ctx context.Context) (expire int, err error) {
	expire, access, err := repo.adminTokenFetch(ctx)

	if err != nil {
		return
	}

	repo.adminMutex.Lock()
	repo.adminAccessToken = access
	repo.adminMutex.Unlock()

	expire -= expire / 10

	return
}

func (repo *repositoryImpl) adminTokenMaintain(ctx context.Context) error {
	expire, err := repo.adminTokenUpdate(ctx)
	if err != nil {
		return err
	}

	timer := time.NewTimer(time.Duration(expire) * time.Second)

	go func() {
		for {
			<-timer.C

			expire, err = repo.adminTokenUpdate(context.Background())
			if err != nil {
				log.WithError(err).Error("updating admin token, retry")
				timer.Reset(time.Second)

				continue
			}

			timer.Reset(time.Duration(expire) * time.Second)
		}
	}()

	return nil
}

func (repo *repositoryImpl) httpRequest(req *http.Request, response interface{}) error {
	if repo.config.Debug {
		bs, err := httputil.DumpRequestOut(req, true)
		if err == nil {
			fmt.Printf("\n>>>\n")

			_, _ = os.Stdout.Write(bs)
		}
	}

	resp, err := repo.config.HTTPClient.Do(req)
	if err != nil {
		return err
	}

	defer func() {
		_ = resp.Body.Close()
	}()

	if response == nil {
		return nil
	}

	if repo.config.Debug {
		var bs []byte

		bs, err = httputil.DumpResponse(resp, true)
		if err == nil {
			fmt.Printf("\n<<<\n")

			_, _ = os.Stdout.Write(bs)

			fmt.Printf("\n---\n")
		}
	}

	data, err := ioutil.ReadAll(resp.Body)
	if err != nil {
		return err
	}

	if resp.StatusCode != http.StatusOK {
		var errresp struct {
			Error string `json:"error"`
		}

		if json.Unmarshal(data, &errresp) == nil {
			return errors.New(errresp.Error)
		}

		return fmt.Errorf("status is not OK: %v, please try again in 30 seconds", resp.Status)
	}

	return json.Unmarshal(data, response)
}

func (repo *repositoryImpl) clientGet(ctx context.Context, path string, response interface{}) error {
	req, err := http.NewRequestWithContext(ctx, http.MethodGet, repo.config.URL+path, nil)
	if err != nil {
		return err
	}

	req.Header.Set("Authorization", repo.basicAuth)

	return repo.httpRequest(req, response)
}

func (repo *repositoryImpl) clientPostForm(
	ctx context.Context,
	url string,
	values url.Values,
	response interface{},
) error {
	req, err := http.NewRequestWithContext(
		ctx,
		http.MethodPost,
		url,
		strings.NewReader(values.Encode()),
	)
	if err != nil {
		return err
	}

	req.Header.Set("Content-Type", "application/x-www-form-urlencoded")

	req.Header.Set("Authorization", repo.basicAuth)

	return repo.httpRequest(req, response)
}

func (repo *repositoryImpl) adminGet(ctx context.Context, path string, response interface{}) error {
	req, err := http.NewRequestWithContext(ctx, http.MethodGet, repo.config.URL+path, nil)
	if err != nil {
		return err
	}

	return repo.httpRetry(func() error {
		req.Header.Set("Authorization", "bearer "+repo.adminToken())

		return repo.httpRequest(req, response)
	})
}

func (repo *repositoryImpl) adminDelete(ctx context.Context, path string, response interface{}) error {
	req, err := http.NewRequestWithContext(ctx, http.MethodDelete, repo.config.URL+path, nil)
	if err != nil {
		return err
	}

	return repo.httpRetry(func() error {
		req.Header.Set("Authorization", "bearer "+repo.adminToken())

		return repo.httpRequest(req, response)
	})
}

func (repo *repositoryImpl) adminSendJSON(
	ctx context.Context,
	method, path string,
	value, response interface{}) error {
	var rdr io.Reader

	if value != nil {
		data, err := json.Marshal(value)
		if err != nil {
			return err
		}

		rdr = bytes.NewReader(data)
	}

	req, err := http.NewRequestWithContext(ctx, method, repo.config.URL+path, rdr)
	if err != nil {
		return err
	}

	req.Header.Set("Content-type", "application/json")

	return repo.httpRetry(func() error {
		req.Header.Set("Authorization", "bearer "+repo.adminToken())

		return repo.httpRequest(req, response)
	})
}

func (repo *repositoryImpl) adminPost(ctx context.Context, path string, value interface{}) error {
	return repo.adminSendJSON(ctx, http.MethodPost, path, value, nil)
}

func (repo *repositoryImpl) adminPut(ctx context.Context, path string, value interface{}, response interface{}) error {
	return repo.adminSendJSON(ctx, http.MethodPut, path, value, response)
}

func (repo *repositoryImpl) httpRetry(f func() error) (err error) {
	for i := 0; i <= httpRetries; i++ {
		err = f()

		if err != nil && err.Error() == err401Unauthorized {
			_, refresherr := repo.adminTokenUpdate(context.Background())
			if refresherr == nil {
				continue
			}
		}

		return
	}

	return
}
